/*
 * @(#)CellRange.java
 *
 * $Id: CellRange.java,v 1.13 2000/12/02 21:44:51 rkc10 Exp $
 *
 * Created on November 5, 2000, 1:19 AM
 */

/** 
 * This class specifies how a range of cells are represented.
 * <p>
 * A range is a continuous rectangular area that can be specified by its
 * upper left corner (minCorner) and lower right corner (maxCorner).
 * 
 * @author Ricky Chin
 * @version $Id: CellRange.java,v 1.13 2000/12/02 21:44:51 rkc10 Exp $
 */
import javax.swing.*;
import java.awt.image.*;
import com.sun.image.codec.jpeg.*;
import java.io.*;
import java.awt.Graphics.*;
import java.awt.Graphics2D;
import java.awt.*;
import javax.swing.table.*;

public class CellRange {
    
    /** Upper left corner of range */
    CellPoint minCorner;

    /** Lower right corner of range */
    CellPoint maxCorner;

    // Table Model
    UMTableModel m_model = null;

    /** Creates new CellRange. In assumes that the array parameters have
     * entries listed in ascending order. In other words, the row array
     * has the row with the smallest number in index 0 and largest
     * row in the last index. The same goes for the column array.
     * @param rows the array range of rows
     * @param cols the array range of columns
     */
    public CellRange(int[] rows, int[] cols) {
        //rows selected are in ascending order
        int minRow = rows[0];
        int maxRow = rows[rows.length - 1];
        
        //columns selected are in ascending order
        int minCol = cols[0];
        int maxCol = cols[cols.length - 1];
		
         minCorner = new CellPoint(minRow, minCol);
         maxCorner = new CellPoint(maxRow, maxCol);
	}
    
    /**
     * This contructor takes x1, x2, y1, y2 and constructs a range.
     * 
     * @param minRow upper left corner row coordinate
     * @param minCol upper left corner col coordinate
     * @param maxRow lower right corner row coordinate
     * @param maxCol lower right corner col coordinate
     */
    public CellRange(int minRow, int maxRow, int minCol, int maxCol) {
	
	minCorner = new CellPoint(minRow, minCol);
	maxCorner = new CellPoint(maxRow, maxCol);
    }

    

    /**
     * This constructor takes two CellPoints to construct a range.
     * 
     * @param ULHCorner upper left corner
     * @param LRHCorner lower right corner
     */
    public CellRange(CellPoint ULHCorner, CellPoint LRHCorner) {
            minCorner = ULHCorner;
            maxCorner = LRHCorner;
    }

    
    
    /**
     * This returns the first row in the range.
     * 
     * @return first row of range
     */
    public int getStartRow() {
        return minCorner.getRow();
    }
    
    /**
     * This returns the last row in the range.
     * 
     * @return last row of range
     */
    public int getEndRow() {
        return maxCorner.getRow();
    }
    
    /**
     * This returns the first column in the range.
     * 
     * @return first column of range
     */
    public int getStartCol() {
        return maxCorner.getCol();
    }
    
    /**
     * This returns the last column in the range.
     * 
     * @return last column of range
     */
    public int getEndCol() {
        return maxCorner.getCol();
    }
    
    /**
     * This returns the number of columns in the range.
     * 
     * @return number of columns in range
     */
    public int getWidth() {
        return getEndCol() - getStartCol() + 1;
    }
    
    /**
     * This returns the number of rows in the range.
     *
     * @return number of rows in range
     */
    public int getHeight() {
        return getEndRow() - getStartRow() + 1;
    }

  

    /** This returns the upper left corner of the range
     * @return the upper left corner of the range
     */
    public CellPoint getminCorner() {
        return minCorner;
    }
    /** This returns the lower right corner of the range
     * @return the lower right corner of the range
     */    
    public CellPoint getmaxCorner() {
        return maxCorner;
    }

    /**
     * This returns a string represetation of the difference of the upper left
     * and lower right corners
     *
     * @return string minCorner - maxCorner
     */
    public String toString() {
	return minCorner.toString()+':'+maxCorner.toString();
    }
    
    /**
     * A new constructor that takes a table model so an image can be created
     *
     */
    public CellRange(int minRow, int maxRow, int minCol, int maxCol,
                     UMTableModel model) {
	
	minCorner = new CellPoint(minRow, minCol);
	maxCorner = new CellPoint(maxRow, maxCol);
        m_model = model;
    }

     /**
     * This returns a represenation of the range as a buffered image,
     * used for copy/paste to image applications, and for printing
     *
     * @return BufferedImage representaion of the range as an image
     */
    public BufferedImage getImage(boolean forPrint)
    {
        
        JTable table = m_model.getTable();
        int curX = 0;
        int curY = 0;
        int h = 0;
        int w = 0;
        int totalH = 0;
        int totalW = 0;
        for (int i = getStartRow(); i <= getEndRow(); i++)
            totalH += table.getRowHeight(i);
        for (int j = getStartCol(); j <= getEndCol(); j++)
            totalW += table.getColumnModel().getColumn(j).getWidth(); 

        BufferedImage img = new BufferedImage(totalW + 1, totalH + 1, 
                                              BufferedImage.TYPE_INT_RGB);
        Graphics2D g = img.createGraphics();
        g.setColor(Color.white);
        g.fillRect(0,0, totalW + 1, totalH + 1);
        for (int i = getStartRow(); i <= getEndRow(); i++)
        {
            h = table.getRowHeight(i);
            for (int j = getStartCol(); j <= getEndCol(); j++)
            {
                w = table.getColumnModel().getColumn(j).getWidth();
                UMCell cell = m_model.getCellAt(i,j);
                Color fColor = cell.getForeground();
                Color bColor = cell.getBackground();
                Font font = cell.getFont();
                
                if (fColor == null)
                    fColor = m_model.getTable().getForeground();
                if (bColor == null)
                    bColor = m_model.getTable().getBackground();
                if (font == null)
                    font = m_model.getTable().getFont();

                g.setColor(bColor);
                g.fillRect(curX, curY, w, h);
                g.setColor(fColor);
                g.setFont(font);
                g.drawString(cell.toString(), curX + 3, curY + h - 3);
                LinesBorder b = (LinesBorder) cell.getBorder();
                if (!forPrint)
                {
                    g.setColor(Color.black);
                    g.drawLine(curX, curY, curX, curY + h);
                    g.drawLine(curX, curY, curX + w, curY);
                    g.drawLine(curX, curY + h, curX + w, curY + h);
                    g.drawLine(curX + w , curY, curX + w, curY + h);
                }
                if (b != null)
                {
                    g.setColor(Color.black);
                    Insets ins = b.getBorderInsets(null);
                    if (ins.top == 2)
                    {
                        //g.drawLine(curX, curY, curX + w, curY); 
                        g.drawLine(curX , curY + 1, 
                                   curX + w , curY + 1);
                    }
                    if (ins.bottom == 2)
                    {
                        //g.drawLine(curX, curY + h, curX + w, curY + h);
                        g.drawLine(curX , curY + h - 1, 
                                   curX + w, curY + h - 1);
                    }
                    if (ins.right == 2)
                    {
                        //g.drawLine(curX + w , curY, curX + w, curY + h);
                        g.drawLine(curX + w - 1, curY, 
                                   curX + w - 1 , curY + h);
                    }
                    if (ins.left == 2)
                    {
                        //g.drawLine(curX, curY, curX, curY + h);
                        g.drawLine(curX + 1, curY, 
                                   curX + 1, curY + h);
                    }
                }
                        
                curX += w;
                
            }
            curY += h;
            curX = 0;
        }
        if (!forPrint)
        {
            g.setColor(Color.black);
            g.drawLine(totalW, 0, totalW, totalH);
            g.drawLine(0,totalH,totalW, totalH);
        }
                
               
        //p.setViewState("CPC");
        /*
        try
        {
            FileOutputStream fos = new FileOutputStream("out.jpg");
            JPEGImageEncoder jpeg = JPEGCodec.createJPEGEncoder(fos);
            jpeg.encode(img);
            fos.close();
        } catch (IOException e){}
        */
        return img;
    }
                
}
